import { ObservableValue } from "./Observable";
import { makePropertyObservable } from "./ObservableObject";
/** _Property decorator_, makes a property observable on every instance, with a read-only value shared across all instances taken from an (earlier OR later) call to `inject`, *until* the property is directly assigned to [requires ES5+ target] [decorator] */
export function injectable(target, key, descriptor) {
    if (!descriptor)
        descriptor = { enumerable: true };
    // capture existing value (e.g. method on prototype)
    var value = descriptor.value;
    delete descriptor.value;
    delete descriptor.writable;
    // use a single observable per class to capture the last injected value
    var injector = ObservableValue.fromValue(value);
    descriptor.get = function () {
        makePropertyObservable(this, key, ObservableValue.fromValue(injector));
        return this[key];
    };
    // setter just initializes an observable with given value
    // (will ignore injection altogether)
    descriptor.set = function (value) {
        makePropertyObservable(this, key);
        this[key] = value;
    };
    // store a reference to the observable injector value
    descriptor.get["*injector"] = injector;
    return descriptor;
}
/** Set given injectable observable properties (decorated with `injectable`) on *all* instances of given class and derived classes to the given values, except for instances where the property has been overwritten directly; may be called multiple times even with the same properties to redefine their injected value; returns an object with the previous injected values (for e.g. overriding an injected function that calls into the previously injected function) */
export function inject(targetClass, spec) {
    var result = {};
    var target = targetClass.prototype;
    for (var name in spec) {
        if (!Object.prototype.hasOwnProperty.call(spec, name))
            continue;
        // get the (prototype) property descriptor and find the injector
        var desc = Object.getOwnPropertyDescriptor(target, name);
        var injector = desc && desc.get && desc.get["*injector"];
        if (!injector)
            throw new TypeError("Not an injected property: " + name);
        // return the current observable injector value and set the new value
        result[name] = injector.getLastValue();
        injector.value = spec[name];
    }
    return result;
}
