import { defer } from "./Defer";
import { Promise } from "./Promise";
/** Helper function that generates a handler function which looks for a given method on given target */
function makeMethodHandler(method, target) {
    if (!target)
        throw new TypeError("Target is " + target);
    return function () {
        var f = target[method];
        if (typeof f !== "function")
            throw new TypeError("Target does not contain method: " + method);
        f.call(target, arguments);
    };
}
/** A static class for a signal that can be used to trigger one or more handlers asynchronously (use `defineSignal` to create classes) */
var Signal = (function () {
    /** Create a signal with given data, ready to be emitted */
    function Signal(data) {
        /** @internal */
        this._emitted = false;
        this._data = data;
    }
    Signal.connect = function (callback, target) {
        // make handler function if given only a method name
        if (typeof callback === "string")
            callback = makeMethodHandler(callback, target);
        // add handler to the list, return a SignalConnection object
        var connection = { disconnected: false };
        connection.disconnect = Signal._connect.call(this, callback, connection);
        return connection;
    };
    Signal.connectOnce = function (callback, target) {
        // make handler function if given only a method name
        if (typeof callback === "string")
            callback = makeMethodHandler(callback, target);
        // add wrapper to the list, call handler only once
        var called = false, un = Signal._connect.call(this, function (data, src) {
            if (!called) {
                called = true;
                un();
                callback(data, src);
            }
        });
    };
    /** @internal Implementation of connect: returns disconnect function [used to speed up connections in other Async classes] */
    Signal._connect = function (callback, signalConnection) {
        var _this = this;
        // add handler to the list
        switch (this._nHandlers++) {
            case 0:
                // call "up" listener to make sure signals get emitted
                this._handler1 = callback;
                this.onHandlerConnected && this.onHandlerConnected();
                break;
            case 1:
                this._handler2 = callback;
                break;
            default:
                (this._handlers || (this._handlers = []))
                    .push(callback);
        }
        return function () {
            // remove handler from the list
            switch (_this._nHandlers) {
                case 0: return;
                case 1:
                    if (_this._handler1 === callback)
                        _this._handler1 = undefined, _this._nHandlers--;
                    break;
                case 2:
                    if (_this._handler1 === callback)
                        _this._handler1 = _this._handler2,
                            _this._handler2 = undefined, _this._nHandlers--;
                    else if (_this._handler2 === callback)
                        _this._handler2 = undefined, _this._nHandlers--;
                    break;
                default:
                    if (_this._handler1 === callback)
                        _this._handler1 = _this._handler2,
                            _this._handler2 = _this._handlers.shift(),
                            _this._nHandlers--;
                    else if (_this._handler2 === callback)
                        _this._handler2 = _this._handlers.shift(),
                            _this._nHandlers--;
                    else
                        for (var i = 0; i < _this._nHandlers; i++) {
                            if (_this._handlers[i] === callback) {
                                _this._handlers.splice(i, 1);
                                _this._nHandlers--;
                                break;
                            }
                        }
            }
            // check new length: call "down" listener if none
            if (!_this._nHandlers && _this.onHandlersDisconnected)
                _this.onHandlersDisconnected();
            // set disconnected flag
            if (signalConnection)
                signalConnection.disconnected = true;
        };
    };
    /** Remove all handlers; returns this */
    Signal.disconnectAll = function () {
        // call "down" callback to allow signals to stop for now
        if (this._nHandlers) {
            this._handler1 = undefined;
            this._handler2 = undefined;
            delete this._handlers;
            this._nHandlers = 0;
            this.onHandlersDisconnected && this.onHandlersDisconnected();
        }
        return this;
    };
    /** Returns true if this signal has any handlers */
    Signal.isConnected = function () {
        return (this._nHandlers > 0);
    };
    /** @internal Invoke all handlers synchronously, without creating a Signal instance at all; exceptions in handlers are NOT caught here */
    Signal.emitSync = function (data) {
        if (this._nHandlers) {
            if (this._handler1)
                this._handler1.call(undefined, data, this);
            if (this._handler2)
                this._handler2.call(undefined, data, this);
            var handlers = this._handlers;
            if (handlers) {
                for (var i = 0, len = handlers.length; i < len; i++)
                    handlers[i] && handlers[i].call(undefined, data, this);
            }
        }
    };
    /** @internal Invoke all handlers synchronously, without creating a Signal instance at all; exceptions in handlers are NOT caught here */
    Signal.emit = function (data) {
        if (this._nHandlers) {
            var args = [data, this];
            if (this._handler1)
                defer(this._handler1, args);
            if (this._handler2)
                defer(this._handler2, args);
            var handlers = this._handlers;
            if (handlers) {
                for (var i = 0, len = handlers.length; i < len; i++)
                    handlers[i] && defer(handlers[i], args);
            }
        }
    };
    /** Invoke all handlers and capture promises of their return values in .results (unless argument is true); works only once; returns this */
    Signal.prototype.emit = function (noResults) {
        if (this._emitted)
            return this;
        this._emitted = true;
        var nHandlers = this.constructor._nHandlers;
        if (!nHandlers)
            return this;
        var handler1 = this.constructor._handler1;
        var handler2 = this.constructor._handler2;
        var handlers = this.constructor._handlers;
        var args = [this._data, this.constructor];
        if (!noResults) {
            // intialize an array of Promises
            var results = this._results = [handler1 ?
                    Promise.defer(handler1, args) :
                    Promise.resolve(undefined)];
            if (nHandlers > 1)
                results.push(handler2 ?
                    Promise.defer(handler2, args) :
                    Promise.resolve(undefined));
            if (nHandlers > 2)
                handlers.forEach(function (f) { return results.push(f ?
                    Promise.defer(f, args) :
                    Promise.resolve(undefined)); });
        }
        else {
            // no need to capture results, schedule plain handlers
            defer(handler1, args);
            if (nHandlers > 1)
                defer(handler2, args);
            if (nHandlers > 2)
                handlers.forEach(function (f) { return f && defer(f, args); });
        }
        if (this._onEmit)
            this._onEmit(true);
        return this;
    };
    /** Invoke given callback(s) with return values of all handlers, or any exception that occurred during execution of all handlers; returns a Promise that resolves to the return value of the callback(s) itself */
    Signal.prototype.then = function (onFulfilled, onRejected) {
        var _this = this;
        if (!this._promise) {
            if (!this._emitted)
                this._promise = new Promise(function (r) { _this._onEmit = r; })
                    .then(function () { return Promise.all(_this.results); });
            else
                this._promise = Promise.all(this.results);
        }
        return this._promise.then(onFulfilled, onRejected);
    };
    /** Catch errors that occur during execution of all handlers; returns a Promise that resolves to the return value of the callback itself */
    Signal.prototype.catch = function (onRejected) {
        return this.then(undefined, onRejected);
    };
    Object.defineProperty(Signal.prototype, "results", {
        /** Array of Promises that resolve to return values of all handlers */
        get: function () {
            return this._results || [];
        },
        enumerable: true,
        configurable: true
    });
    return Signal;
}());
export { Signal };
/** @internal Number of handlers */
Signal._nHandlers = 0;
export function defineSignal(Base, props) {
    if (Base === void 0) { Base = Signal; }
    var Result = function Signal(_token) {
        if (!(this instanceof Result)) {
            // called as a function, not a constructor: emit directly
            Result.emit.apply(Result, arguments);
        }
        else {
            // called as a constructor, just call base constructor
            Base.apply(this, arguments);
        }
    };
    // copy static methods/members and apply prototype
    _object_assign(Result, Base);
    _object_assign(Result, props);
    Result.prototype = Object.create(Base.prototype);
    Result.prototype.constructor = Result;
    // clear all handlers for this class object and return the class
    Result._nHandlers = 0;
    return Result;
}
// Simple polyfill for Object.assign
var _object_assign = Object.assign;
if (!_object_assign) {
    _object_assign = function (target, source) {
        if (source) {
            for (var id in source) {
                if (Object.prototype.hasOwnProperty.call(source, id))
                    target[id] = source[id];
            }
        }
        return target;
    };
}
/** Signal that is triggered with exceptions that were unhandled during async execution; a custom handler may be added here, the default handler just writes a warning message to the console; to disable this behavior for specific errors, set error.message to a blank string */
export var UnhandledException = defineSignal();
// hook up default behavior:
UnhandledException._connect(function (error) {
    try {
        error && error.message && console && (console.warn || console.log)("Unhandled exception in asynchronous code - ", error);
    }
    catch (all) { }
});
