"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var Defer_1 = require("./Defer");
var Promise_1 = require("./Promise");
var Signal_1 = require("./Signal");
var ObservableObject_1 = require("./ObservableObject");
var ObservableArray_1 = require("./ObservableArray");
/** Watched observable currently evaluating, if any; used to connect dependencies through signals */
var currentWatchedEvaling = null;
/** Unwatched observable currently evaluating, if any; used to collect dependencies for dirty checking */
var currentUnwatchedEvaling = null;
/** @internal Throws an error if currently running wached observable getter */
function assertUnobserved() {
    if (currentWatchedEvaling || currentUnwatchedEvaling)
        throw new Error("Observable getter must be a pure function, " +
            "cannot create or set other observables");
}
exports.assertUnobserved = assertUnobserved;
/** Represents an observable value (with the value itself in `.value`) */
var ObservableValue = (function () {
    /** Create a new (single) observed value using optional getter and setter functions; note that getters *should* be pure functions without side effects, creating or setting other observables from the getter will result in an error; getter is not called immediately but only when being subscribed to or when obtaining the value itself (from `.value`) */
    function ObservableValue(getter, setter) {
        /** @internal Unique ID used to quickly index dependants of observables */
        this._uid = String(++ObservableValue.UID);
        /** @internal Counter that is incremented every time the value becomes unstable (i.e. value set directly, dependency changed, etc.) */
        this._dirtyIdx = 0;
        assertUnobserved();
        // hold off on evaluating until value getter is called
        if (getter) {
            this._getter = getter;
            this._dirtyIdx++;
        }
        this._setter = setter;
    }
    /** Encapsulate given value as an ObservableValue */
    ObservableValue.fromValue = function (value) {
        var result = new ObservableValue();
        result.value = value;
        return result;
    };
    /** Encapsulate given promised value as an ObservableValue; the observed value will start out as `undefined` but then changes to the `Promise` result when resolved; or if promise was rejected, hangs on to the error and throws it when retrieving the observable value */
    ObservableValue.fromPromise = function (valuePromise) {
        var result = new ObservableValue();
        if (valuePromise instanceof Promise_1.Promise) {
            // shortcut promise status to avoid async
            var status = valuePromise._getStatus();
            if (status) {
                if (status.rejected) {
                    // set value to "error" to trigger change signal
                    result.value = "error";
                    result._error = status.error;
                    result._ro = true;
                    return result;
                }
                else {
                    // set value to resolved promise value
                    result.value = status.value;
                    result._ro = true;
                    return result;
                }
            }
        }
        valuePromise.then(function (value) {
            // set value to resolved promise value
            result.value = value;
            result._ro = true;
        }, function (error) {
            // set value to "error" to trigger change signal
            result.value = "error";
            result._error = error;
            result._ro = true;
        });
        return result;
    };
    /** Set a getter function for the observable value, which should return a current value, or another ObservableValue instance, or set `.value` directly; note that getters *should* be pure functions without side effects, creating or setting other observables from the getter will result in an error; the getter is only invoked (asynchronously) if a value had been set previously and needs to be updated; reading `.value` from the getter results in the value previously set, no recursion occurs; returns this */
    ObservableValue.prototype.getter = function (f) {
        assertUnobserved();
        this._getter = f;
        // remove current static value and re-eval if needed
        if (Object.prototype.hasOwnProperty.call(this, "_val")) {
            if (f)
                delete this._val;
            this._asyncEval();
        }
        return this;
    };
    /** Set a setter function for the observable value; setting `.value` directly from the setter results in changing the underlying observable value (which can also be read by the getter, or the setter itself), no recursion occurs; returns this */
    ObservableValue.prototype.setter = function (f) {
        assertUnobserved();
        this._setter = f;
        return this;
    };
    Object.defineProperty(ObservableValue.prototype, "value", {
        /** Observable value, (re-) evaluated only if necessary; when set to an `ObservableValue` instance, this property returns that instance's value, until this property is set to another value (except if the `ObservableValue` has a setter, which is called first); plain Array values are turned into `ObservableArray` instances, and plain Object instances into `ObservableObject` instances --- unless `.shallow` is set to true */
        get: function () {
            if (this._error)
                throw this._error;
            if (this._getting)
                return this._val;
            if (!this._setting && !this._ro) {
                // register as dependency to previously evaluating observable
                if (currentWatchedEvaling)
                    currentWatchedEvaling._addConnection(this);
                else if (currentUnwatchedEvaling)
                    currentUnwatchedEvaling._addCheck(this);
                // re-evaluate if necessary:
                if (this._needsEval()) {
                    this._oldConnections = this._depConnections;
                    var hasValue = this.hasOwnProperty("_val");
                    var oldValue = this._val;
                    // set this instance as currently evaluating to find dependencies
                    var prevWatched = currentWatchedEvaling, prevUnwatched = currentUnwatchedEvaling;
                    if (this._watched) {
                        currentWatchedEvaling = this;
                        this._depConnections = {};
                    }
                    else {
                        currentUnwatchedEvaling = this;
                        this._depChecks = [];
                    }
                    this._getting = true;
                    this._direct = false;
                    try {
                        // store getter result or let getter set value directly
                        var result = this._getter.call(undefined);
                        if (!(result === undefined && this._direct))
                            this._val = result;
                    }
                    finally {
                        this._getting = false;
                        // stop solliciting dependencies
                        currentWatchedEvaling = prevWatched;
                        currentUnwatchedEvaling = prevUnwatched;
                    }
                    // mark value as up to date
                    this._valIdx = this._dirtyIdx;
                    // unsubscribe from old dependencies that are no longer used
                    var oldConnections = this._oldConnections;
                    delete this._oldConnections;
                    if (this._depConnections) {
                        for (var uid in oldConnections)
                            if (!this._depConnections[uid])
                                oldConnections[uid].dis();
                    }
                    // emit signal(s) if changed
                    if (hasValue && this._val !== oldValue) {
                        this._auxSignalers && this._auxSignalers.forEach(function (f) { return f(); });
                        this._Signal && this._Signal.emitSync(this._val);
                    }
                }
            }
            // return new value or proxy other ObservableValue's value
            var value = this._val;
            if (!this.shallow)
                while (value instanceof ObservableValue)
                    value = value.value;
            return value;
        },
        set: function (value) {
            if (!this._getting && !this._setting)
                assertUnobserved();
            // convert plain arrays and objects (unless about to run setter)
            if (!this.shallow && (!this._setter || this._setting)) {
                // make ObservableArray instances out of plain Arrays
                if (value instanceof Array && value.constructor === Array) {
                    value = ObservableArray_1.ObservableArray.fromArray(value);
                }
                // make ObservableObject instances out of plain Objects
                if (value instanceof Object && value.constructor === Object) {
                    value = ObservableObject_1.makeObjectObservable(value);
                }
            }
            // check what to do with the new value:
            if (this._getting) {
                // set value directly, no signals but mark as changed
                this._val = value;
                this._direct = true;
                this._valIdx = ++this._dirtyIdx;
            }
            else if (this._getter && !this._setter || this._ro) {
                // cannot set value directly outside of getter function
                // (or value is resolved Promise)
                throw new Error("Cannot set value");
            }
            else if ((!this._setter || this._setting) &&
                (this._val instanceof ObservableValue) &&
                this._val._setter) {
                // proxy through observable value setter
                this._val.value = value;
            }
            else if (this._setter && !this._setting) {
                // invoke setter
                this._setting = true;
                try {
                    this._setter.call(undefined, value);
                }
                finally {
                    this._setting = false;
                }
            }
            else if (value !== this._val) {
                // mark value as changed (but not dirty)
                this._valIdx = ++this._dirtyIdx;
                // set value and emit signal(s)
                this._val = value;
                this._auxSignalers && this._auxSignalers.forEach(function (f) { return f(); });
                this._Signal && this._Signal.emitSync(value);
            }
        },
        enumerable: true,
        configurable: true
    });
    /** Returns .value (observable if used within an observable getter) */
    ObservableValue.prototype.valueOf = function () { return this.value; };
    /** Returns .value as a string (observable if used within an observable getter) */
    ObservableValue.prototype.toString = function () { return String(this.value); };
    /** Returns last value set, does not re-evaluate and/or add dependency */
    ObservableValue.prototype.getLastValue = function () {
        // return value or proxy other ObservableValue's value
        var value = this._val;
        while (value instanceof ObservableValue)
            value = value.getLastValue();
        return value;
    };
    /** Transform this observable value using given function, into a new ObservableValue instance; note that the transformation function is not necessarily invoked after each value change, if the new observable is not subscribed to and/or multiple changes occur (asynchronously) before the transformation is evaluated; observables used by the given function itself are not automatically subscribed to (wrap given function in observe(...) to observe dependencies as well) */
    ObservableValue.prototype.map = function (callback) {
        var _this = this;
        var lastValue = {};
        var mapped = new ObservableValue(function () {
            var value = _this.value;
            if (value === lastValue)
                return mapped.value;
            return unobserved(callback, lastValue = value);
        });
        return mapped;
    };
    ObservableValue.prototype.subscribe = function (callback) {
        var _this = this;
        (this._Signal || this._sig())._connect(callback);
        // listen for signal emissions while initializing value if needed
        var current = this._val, emitted = false, f;
        if (this._getter || this._val instanceof ObservableValue) {
            if (!this._auxSignalers)
                this._auxSignalers = [];
            this._auxSignalers.push(f = function () { emitted = true; });
            unobserved(function () { current = _this.value; });
            if (this._auxSignalers.length === 1)
                delete this._auxSignalers;
            else
                this._auxSignalers = this._auxSignalers.filter(function (v) { return v !== f; });
        }
        // if not already emitted signal, invoke callback for current value
        if (!emitted)
            callback && Defer_1.defer(callback, [current]);
        return this;
    };
    /** Start listening for changes to this observable value and all of its dependencies asynchronously, and return a promise for the next (different) value of the observable; stops subscribing immediately after the promise is resolved */
    ObservableValue.prototype.next = function () {
        var _this = this;
        var result = new Promise_1.Promise(function (resolve) {
            (_this._Signal || _this._sig()).connectOnce(function (v) { resolve(v); });
        });
        // initialize value if needed
        if (this._getter || this._val instanceof ObservableValue)
            unobserved(function () { _this.value; });
        return result;
    };
    Object.defineProperty(ObservableValue.prototype, "subscribed", {
        /** True if this observable value is currently subscribed to, either directly or from dependent observables */
        get: function () { return this._watched; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ObservableValue.prototype, "writable", {
        /** True if this observable value is writable (not only a getter defined) */
        get: function () { return !this._getter || !!this._setter; },
        enumerable: true,
        configurable: true
    });
    /** @internal Add a signal to emit when this value changes, but do not watch for changes directly (i.e. does not connect to signal and start watching); used e.g. to emit Observable.PropertyChange */
    ObservableValue.prototype.emitOnChange = function (signal, data) {
        if (!this._auxSignalers)
            this._auxSignalers = [];
        this._auxSignalers.push(function () { signal(data); });
    };
    /** Clear value and remove getter/setter functions, unsubscribe from dependencies */
    ObservableValue.prototype.clear = function () {
        this.getter(null).setter(null);
    };
    /** Manually invoke the getter function synchronously, if any; can be used only if not currently evaluating another observable value (otherwise, wrap in a call to `unobserved`); calling this method is normally not necessary, and should only be used if external factors outside of observables change in such a way that the result of the getter function changes; if so, then subsequent retrieval of `.value` results in the new value, and dependent observable values are scheduled to update automatically (asynchronously) */
    ObservableValue.prototype.update = function () {
        assertUnobserved();
        if (this._getter) {
            this._dirtyIdx++;
            this.value;
        }
    };
    /** @internal Returns true if value needs to be reevaluated */
    ObservableValue.prototype._needsEval = function () {
        if (!this._getter)
            return false;
        if ((this._valIdx || 0) !== this._dirtyIdx)
            return true;
        // if value is watched, then dependencies are watched too
        // and are up to date if value is up to date (above check)
        if (this._watched)
            return false;
        // check if dependencies (may) have changed since last evaluation
        return !this._depChecks || this._depChecks.some(function (f) { return f({}); });
    };
    /** @internal Mark as dirty (if not already dirty) and schedule re-evaluation of value if still watched */
    ObservableValue.prototype._asyncEval = function () {
        var _this = this;
        if ((this._valIdx || 0) === this._dirtyIdx && !this._getting) {
            this._dirtyIdx++;
            Defer_1.defer(function () {
                if (_this._watched && _this._valIdx !== _this._dirtyIdx)
                    _this.value;
            });
        }
    };
    /** @internal Register a watched dependency of this value */
    ObservableValue.prototype._addConnection = function (other) {
        if (other === this)
            return;
        if (!this._depConnections)
            this._depConnections = {};
        // check if connection already existed
        var old;
        if (this._oldConnections && (old = this._oldConnections[other._uid])) {
            // use existing object
            this._depConnections[other._uid] = old;
        }
        else {
            // add handler to mark this value as dirty whenever other value
            // changes (not async since signal is always emitted synchronously,
            // see above); _connect returns disconnect method
            var dis = (other._Signal || other._sig())._connect(this._asyncEval.bind(this));
            // store new dependency object
            this._depConnections[other._uid] = { dep: other, dis: dis };
        }
    };
    /** @internal Register an unwatched dependency of this value (checked for changes every time when getting the unwatched observable value) */
    ObservableValue.prototype._addCheck = function (other) {
        var _this = this;
        if (other === this)
            return;
        // maintain a function to check if the other value has changed
        // (different _dirtyIdx or recurse down dependency tree)
        var lastDirtyIdx = other._dirtyIdx;
        (this._depChecks || (this._depChecks = [])).push(function (seen) {
            if (seen[_this._uid])
                return true;
            seen[_this._uid] = true;
            return (other._dirtyIdx !== lastDirtyIdx ||
                !other._depChecks || other._depChecks.some(function (f) { return f(seen); }));
        });
    };
    /** @internal Disconnect existing dependencies */
    ObservableValue.prototype._removeDependencies = function () {
        delete this._depChecks;
        var connections = this._depConnections;
        delete this._depConnections;
        for (var uid in connections)
            connections[uid].dis();
    };
    /** @internal Instantiate and return the value change signal */
    ObservableValue.prototype._sig = function () {
        return (this._Signal = Signal_1.defineSignal(ObservableValue.ValueChangeSignal, { ref: this }));
    };
    return ObservableValue;
}());
/** @internal */
ObservableValue.UID = 0;
/** @internal Signal class with initializer */
ObservableValue.ValueChangeSignal = (function (_super) {
    __extends(class_1, _super);
    function class_1() {
        return _super.apply(this, arguments) || this;
    }
    /** Called when a first handler is connected */
    class_1.onHandlerConnected = function () {
        var ref = this.ref;
        ref._watched = true;
        if (!ref._depChecks || ref._depChecks.length) {
            // not sure about dependencies, force re-eval
            ref._removeDependencies();
            ref._dirtyIdx++;
        }
        else {
            // no dependencies last time, no need to check
            ref._removeDependencies();
        }
    };
    /** Called when all handlers have been disconnected */
    class_1.onHandlersDisconnected = function () {
        // all handlers have been disconnected:
        this.ref._watched = false;
        this.ref._removeDependencies();
    };
    return class_1;
}(Signal_1.Signal));
exports.ObservableValue = ObservableValue;
function observe(v) {
    return (typeof v === "function") ? new ObservableValue(v) :
        (v && typeof v === "object" && typeof v.then === "function") ?
            ObservableValue.fromPromise(v) :
            (v instanceof Array) ?
                ObservableArray_1.ObservableArray.fromArray(v) :
                (v instanceof Object) ?
                    ObservableObject_1.makeObjectObservable(v) :
                    (function () { throw new TypeError(); })();
}
exports.observe = observe;
function unobserved(f) {
    var prevWatched = currentWatchedEvaling;
    var prevUnwatched = currentUnwatchedEvaling;
    try {
        currentWatchedEvaling = null;
        currentUnwatchedEvaling = null;
        return f.apply(undefined, Array.prototype.slice.call(arguments, 1));
    }
    finally {
        currentWatchedEvaling = prevWatched;
        currentUnwatchedEvaling = prevUnwatched;
    }
}
exports.unobserved = unobserved;
