"use strict";
var Observable_1 = require("./Observable");
var ObservableObject_1 = require("./ObservableObject");
;
/** Encapsulates `Array` with observable properties; the result works exactly like a regular array, but setting elements outside the bounds of the array (>= length) does NOT work: length must be set first */
var ObservableArray = (function () {
    /** Create a new empty `ObservableArray` instance */
    function ObservableArray() {
        var _this = this;
        Observable_1.assertUnobserved();
        var _length = Observable_1.ObservableValue.fromValue(0);
        Object.defineProperty(this, "length", {
            get: function () { return _length.value; },
            set: function (value) {
                // make new numeric properties observable
                var currentLength = _length.getLastValue() || 0;
                while (currentLength < value)
                    ObservableObject_1.makePropertyObservable(_this, currentLength++);
                // clear removed properties
                for (var l = value; l < currentLength; l++)
                    ObservableObject_1.deleteObservableProperty(_this, l);
                _length.value = value;
            },
            configurable: true
        });
    }
    /** Create an `ObservableArray` out of a regular Array */
    ObservableArray.fromArray = function (array) {
        var result = new ObservableArray();
        for (var i = array.length - 1; i >= 0; i--)
            result[i] = array[i];
        result.length = array.length;
        return result;
    };
    /** Create an `ObservableArray` that takes array elements from the array in given `ObservableValue`, or the value itself as a single element if it is not an array, or an empty array if the value is null or undefined; changes in array elements and/or observable value are reflected asynchronously */
    ObservableArray.fromObservableValue = function (observableValue) {
        var result = new ObservableArray();
        // proxy length and all numeric properties in result
        var _length = 0;
        var makeMap = function (i) { return new Observable_1.ObservableValue(function () {
            var value = observableValue.value;
            if (value instanceof Array)
                return value[i];
            else if (i === 0 && value !== null)
                return value;
            else
                return undefined;
        }); };
        Object.defineProperty(result, "length", {
            get: function () {
                var value = observableValue.value;
                var newLength = (value instanceof Array) ? value.length :
                    (value === undefined || value === null) ? 0 : 1;
                if (newLength > _length) {
                    Observable_1.unobserved(function () {
                        // make new numeric properties observable
                        for (var l = _length; l < newLength; l++)
                            ObservableObject_1.makePropertyObservable(result, l, makeMap(l));
                    });
                }
                else {
                    Observable_1.unobserved(function () {
                        // clear removed properties
                        for (var l = newLength; l < _length; l++)
                            ObservableObject_1.deleteObservableProperty(result, l);
                    });
                }
                return _length = newLength;
            },
            set: function () {
                throw new Error("Cannot modify length of mapped array");
            }
        });
        // initialize all values
        Observable_1.unobserved(function () { return result.length; });
        return result;
    };
    /** Create a read-only `ObservableArray` with each value of the original array mapped to the result of the given getter function; observable values used in the map function are not observed (like `ObservableValue#map`); the getter is called for combinations of value and index (i.e. deleting a value in the middle of the source array will trigger changes for all elements after it; if the index is not important then use `.mapAsyncValues` instead); the resulting array length changes along with the original array length */
    ObservableArray.prototype.mapAsync = function (callback, thisArg) {
        var _this = this;
        var result = new ObservableArray();
        // keep length in an observable together just like the properties to avoid
        // unexpected differences in change propagation timings
        var observableLength = new Observable_1.ObservableValue(function () { return _this.length; });
        // proxy all numeric properties in result
        var _length = 0;
        var makeMap = function (i) {
            var lastValue = {};
            var mapped = new Observable_1.ObservableValue(function () {
                var value = _this[i];
                if (value === lastValue)
                    return mapped.value;
                return Observable_1.unobserved(function (value) {
                    return (i >= 0 && i < _this.length) ?
                        callback.call(thisArg || result, value, i, _this) :
                        undefined;
                }, lastValue = value);
            });
            return mapped;
        };
        Object.defineProperty(result, "length", {
            get: function () {
                var newLength = observableLength.value;
                if (newLength > _length) {
                    Observable_1.unobserved(function () {
                        // make new numeric properties observable
                        for (var l = _length; l < newLength; l++)
                            ObservableObject_1.makePropertyObservable(result, l, makeMap(l));
                    });
                }
                else if (newLength < _length) {
                    Observable_1.unobserved(function () {
                        // clear removed properties
                        for (var l = newLength; l < _length; l++)
                            ObservableObject_1.deleteObservableProperty(result, l);
                    });
                }
                return _length = newLength;
            },
            set: function () {
                throw new Error("Cannot modify length of mapped array");
            }
        });
        // initialize all values
        Observable_1.unobserved(function () { return result.length; });
        return result;
    };
    /** Create a read-only ObservableArray with each value of the original array mapped to the result of the given function; observable values used in the map function are not observed (like ObservableValue map method); the resulting array length changes along with the original array length; this method is slightly more expensive than mapAsync for larger arrays, especially in non-ES6 environments, but is overall more efficient because it avoids unnecessary callbacks when subscribed to */
    ObservableArray.prototype.mapAsyncValues = function (callbackfn) {
        var _this = this;
        // keep track of previously seen values in a Map (with "semi-polyfill")
        var _get, _add;
        var _del;
        var _forEach;
        if (typeof Map === "function") {
            // use native Map implementation
            var map = new Map();
            _get = map.get.bind(map), _add = map.set.bind(map);
            _forEach = map.forEach.bind(map);
            _del = map.delete.bind(map);
        }
        else {
            // use minimal approximation of Map methods
            var ownMap = [];
            _add = function (v, o) { ownMap.push({ v: v, o: o }); };
            _get = function (v) {
                var result;
                ownMap.some(function (x) { return (x.v === v ? (result = x.o, true) : false); });
                return result;
            };
            _del = function (v) {
                ownMap.some(function (x, i) { return (x.v === v ? !!ownMap.splice(i, 1) : false); });
            };
            _forEach = function (f) {
                ownMap.slice(0).forEach(function (x) { return f(x.o, x.v); });
            };
        }
        // keep track of all current inputs to be able to prune the map
        // (really only works when entire array is subscribed to)
        var currentInputs = [];
        var timer = -1;
        var deferPrune = function () {
            if (timer >= 0)
                clearTimeout(timer);
            var myTimer = timer = setTimeout(function () {
                if (myTimer === timer) {
                    timer = -1;
                    var len = _this.length;
                    if (currentInputs.length > len)
                        currentInputs.length = len;
                    _forEach(function (o, v) {
                        currentInputs.indexOf(v) < 0 && _del(v);
                    });
                }
            }, 1);
        };
        // prune the map when the array gets shorter
        new Observable_1.ObservableValue(function () { return _this.length; }).subscribe(function (len) {
            if (currentInputs.length > len)
                deferPrune();
        });
        // map input array to another observable array
        var ownNaN = {}, ownUndefined = {};
        return this.mapAsync(function (v, i) {
            var defv = v === undefined ? ownUndefined :
                (typeof v === "number" && isNaN(v)) ? ownNaN : v;
            if (currentInputs[i] !== undefined && currentInputs[i] !== defv)
                deferPrune();
            currentInputs[i] = defv;
            // return last result for same input value if found
            var found = _get(defv);
            if (found !== undefined)
                return found;
            // get result for current input value and keep it for later
            var o = callbackfn(v);
            _add(defv, o);
            return o;
        });
    };
    /** Represent observable arrays as regular JSON arrays */
    ObservableArray.prototype.toJSON = function () {
        return this.slice(0);
    };
    return ObservableArray;
}());
exports.ObservableArray = ObservableArray;
var _mapAsync = ObservableArray.prototype.mapAsync;
var _mapAsyncValues = ObservableArray.prototype.mapAsyncValues;
var _toJSON = ObservableArray.prototype.toJSON;
ObservableArray.prototype = new Array;
ObservableArray.prototype.constructor = ObservableArray;
ObservableArray.prototype.mapAsync = _mapAsync;
ObservableArray.prototype.mapAsyncValues = _mapAsyncValues;
ObservableArray.prototype.toJSON = _toJSON;
// these functions should not depend on .length but would if not wrapped
var _arrayPush = ObservableArray.prototype.push;
ObservableArray.prototype.push = function () {
    var _this = this;
    var items = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        items[_i] = arguments[_i];
    }
    return Observable_1.unobserved(function () { return _arrayPush.apply(_this, items); });
};
var _arrayUnshift = ObservableArray.prototype.unshift;
ObservableArray.prototype.unshift = function () {
    var _this = this;
    var items = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        items[_i] = arguments[_i];
    }
    return Observable_1.unobserved(function () { return _arrayUnshift.apply(_this, items); });
};
// override concat, which otherwise would see ObservableArrays as values:
ObservableArray.prototype.concat = function () {
    var items = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        items[_i] = arguments[_i];
    }
    return Array.prototype.concat.apply(this.slice(0), items.map(function (item) { return (item instanceof ObservableArray) ? item.slice(0) : item; }));
};
/** Returns an ObservableArray instance based on the result of given function: if the returned value is an array, then all elements are copied to the result; if the returned value is an observable array, all elements are proxied; if the returned value is not an array, the result contains this value as a single element; if the returned value is null or undefined, the result will be an empty array; changes in values/elements are reflected asynchronously */
function observeArray(f) {
    return ObservableArray.fromObservableValue(new Observable_1.ObservableValue(f));
}
exports.observeArray = observeArray;
