"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var Signal_1 = require("./Signal");
var Observable_1 = require("./Observable");
/** Represents an object with (some) observable members; may be extended into a derived class, or use `makeObjectObservable` function to mix into any object [requires ES5+ target] */
var ObservableObject = (function () {
    /** Create an object with (some) observable members */
    function ObservableObject() {
        Observable_1.assertUnobserved();
        _addSignalGetter(this);
    }
    /** Returns true if property with given name is observable */
    ObservableObject.prototype.hasObservableProperty = function (name) {
        return isObservableProperty(this, name);
    };
    return ObservableObject;
}());
exports.ObservableObject = ObservableObject;
(function (ObservableObject) {
    /** Base class for signals that are emitted when any observed property is modified on an ObservableObject instance, with the name of the property as a parameter */
    var PropertyChangeSignal = (function (_super) {
        __extends(PropertyChangeSignal, _super);
        function PropertyChangeSignal() {
            return _super.apply(this, arguments) || this;
        }
        return PropertyChangeSignal;
    }(Signal_1.Signal));
    ObservableObject.PropertyChangeSignal = PropertyChangeSignal;
})(ObservableObject = exports.ObservableObject || (exports.ObservableObject = {}));
exports.ObservableObject = ObservableObject;
/** Returns a new `ObservableObject` with properties copied from given object; `ObservableValue` instances are referenced directly (including getter and setter), Array properties are turned into `ObservableArray` instances, and plain Object instances into `ObservableObject` instances; also seals the new instance if the object parameter was sealed, and freezes it if the object parameter was frozen*/
function makeObjectObservable(obj) {
    var result = new ObservableObject();
    if (obj instanceof Object) {
        // define non-configurable properties with getters and setters
        var specIsObservable = (obj instanceof ObservableObject);
        var desc;
        for (var member in obj) {
            if ((specIsObservable ?
                obj.hasObservableProperty(member) :
                Object.prototype.hasOwnProperty.call(obj, member)) &&
                (desc = Object.getOwnPropertyDescriptor(obj, member))) {
                _defineObservableProperty(result, member, desc.value, desc.get, desc.set);
            }
        }
        // mixin method and signal
        Object.defineProperty(result, "hasObservableProperty", {
            configurable: false,
            writable: false,
            enumerable: false,
            value: ObservableObject.prototype.hasObservableProperty
        });
        _addSignalGetter(result);
        // copy object status
        if (Object.isFrozen(obj)) {
            // go through all properties to initialize hidden observable first
            for (var member in obj)
                result[member];
            Object.freeze(result);
        }
        else if (Object.isSealed(obj)) {
            // go through all properties to initialize hidden observable first
            for (var member in obj)
                result[member];
            Object.seal(result);
        }
    }
    return result;
}
exports.makeObjectObservable = makeObjectObservable;
/** Make the property with given name observable; use given observable value instance, if any, otherwise use a new observable value instance that holds the current property value and/or uses the current property getter and setter [requires ES5+ target] */
function makePropertyObservable(obj, member, instance) {
    var propertyName = String(member);
    if (instance instanceof Observable_1.ObservableValue) {
        // use given instance
        _defineObservableProperty(obj, propertyName, instance);
    }
    else {
        // use property descriptor, or otherwise just the value
        var desc = Object.getOwnPropertyDescriptor(obj, propertyName);
        _defineObservableProperty(obj, propertyName, desc ? desc.value : obj[propertyName], desc && desc.get, desc && desc.set);
    }
}
exports.makePropertyObservable = makePropertyObservable;
/** Delete the property with given name and clear its observable value */
function deleteObservableProperty(obj, member) {
    var hiddenMember = _getHiddenName(member);
    var observable = obj[hiddenMember];
    if (observable && observable.subscribed && observable.writable)
        observable.value = undefined;
    delete obj[member];
    delete obj[hiddenMember];
}
exports.deleteObservableProperty = deleteObservableProperty;
/** Returns true if property with given name is defined and is observable */
function isObservableProperty(obj, member) {
    var hiddenMember = _getHiddenName(member);
    if (Object.prototype.hasOwnProperty.call(obj, hiddenMember))
        return true;
    var ppDescriptor = _getProtoPropDesc(obj, member);
    if (ppDescriptor && ppDescriptor.get && ppDescriptor.get["*observable-getter"])
        return true;
    return false;
}
exports.isObservableProperty = isObservableProperty;
// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
/** Helper function to add a getter for the `ObservableObject.PropertyChange` signal to any object */
function _addSignalGetter(obj) {
    if (!Object.prototype.hasOwnProperty.call(obj, "PropertyChange")) {
        var Signal;
        Object.defineProperty(obj, "PropertyChange", {
            configurable: false,
            enumerable: false,
            get: function () { return (Signal || (Signal = Signal_1.defineSignal(ObservableObject.PropertyChangeSignal, { target: obj }))); }
        });
    }
}
/** Helper function to get a prototype's property descriptor */
function _getProtoPropDesc(obj, member) {
    var proto = Object.getPrototypeOf(obj);
    return proto ?
        Object.getOwnPropertyDescriptor(proto, String(member)) ||
            _getProtoPropDesc(proto, member) :
        null;
}
/** Helper function to add a getter and setter for a single property */
function _defineObservableProperty(obj, member, initial, getter, setter) {
    var hiddenMember = _getHiddenName(member);
    // clear existing observable value, if any, or define hidden property
    if (Object.prototype.hasOwnProperty.call(obj, hiddenMember))
        deleteObservableProperty(obj, member);
    else {
        Object.defineProperty(obj, hiddenMember, {
            configurable: true,
            enumerable: false,
            value: false,
            writable: true
        });
    }
    // define observable property itself
    var descriptor = { enumerable: true, configurable: true };
    _makeObservablePropertyDef(descriptor, member, getter, setter);
    Object.defineProperty(obj, member, descriptor);
    // store observable value instance, if referenced directly
    if (initial instanceof Observable_1.ObservableValue)
        obj[hiddenMember] = initial;
    else
        obj[member] = initial;
}
/** Helper function to get a property definition for an observable property */
function _makeObservablePropertyDef(descriptor, member, getter, setter, getFilter, setFilter, shallow) {
    var hiddenMember = _getHiddenName(member);
    // helper function that returns ObservableValue for a member
    function getObservableValue(obj) {
        if (obj[hiddenMember])
            return obj[hiddenMember];
        // create new observable value
        return obj[hiddenMember] = Observable_1.unobserved(function () {
            var o = new Observable_1.ObservableValue();
            if (shallow)
                o.shallow = true;
            if (getter)
                o.getter(getter.bind(obj));
            if (setter)
                o.setter(setter.bind(obj));
            // emit PropertyChange for Observable object properties
            var propertyChangeSignal = obj.PropertyChange;
            if (typeof propertyChangeSignal === "function")
                o.emitOnChange(propertyChangeSignal, member);
            return o;
        });
    }
    // save getter with flag
    descriptor.get = getFilter ?
        function () {
            return getFilter(getObservableValue(this).value);
        } :
        function () {
            return getObservableValue(this).value;
        };
    descriptor.get["*observable-getter"] = true;
    // save setter
    descriptor.set = setFilter ?
        function (v) {
            return getObservableValue(this).value = setFilter(v);
        } :
        function (v) {
            return getObservableValue(this).value = v;
        };
}
function _getHiddenName(name) {
    return "*observable:" + name;
}
function observable(target, key, descriptor, getFilter, setFilter, shallow) {
    if (!descriptor)
        descriptor = { enumerable: true };
    _makeObservablePropertyDef(descriptor, key, descriptor.get, descriptor.set, getFilter, setFilter, shallow);
    return descriptor;
}
exports.observable = observable;
/** _Property/accessor decorator_, makes a property observable on every instance and converts values to strings when read (empty string for null/undefined/NaN) [requires ES5+ target] [decorator] */
function observable_string(target, key, descriptor) {
    return observable(target, key, descriptor, function (value) {
        return (value || value === 0 || value === false) ? String(value) : "";
    });
}
exports.observable_string = observable_string;
/** _Property/accessor decorator_, makes a property observable on every instance and converts values to numbers using Number(...) when read [requires ES5+ target] [decorator] */
function observable_number(target, key, descriptor) {
    return observable(target, key, descriptor, function (value) { return Number(value); });
}
exports.observable_number = observable_number;
/** _Property/accessor decorator_, makes a property observable on every instance, as a shallow observable value (i.e. does not read values from observable values assigned to this property; and does not convert arrays to observable arrays nor objects to observable objects) [requires ES5+ target] [decorator] */
function observable_shallow(target, key, descriptor) {
    return observable(target, key, descriptor, undefined, undefined, true);
}
exports.observable_shallow = observable_shallow;
/** _Property/accessor decorator_, makes a property observable on every instance and enforces that values are not undefined or null when read (throws TypeError) and that undefined/null values cannot be set (also throws TypeError) [requires ES5+ target] [decorator] */
function observable_not_null(target, key, descriptor) {
    return observable(target, key, descriptor, function (value) {
        if (value === undefined || value === null)
            throw new TypeError(key + " is " + value);
        return value;
    }, function (value) {
        if (value === undefined || value === null)
            throw new TypeError(key + " cannot be " + value);
        return value;
    });
}
exports.observable_not_null = observable_not_null;
/** @internal Alias for backwards compatibility */
function observable_nonNull() {
    return observable_not_null.call(undefined, arguments);
}
exports.observable_nonNull = observable_nonNull;
/** _Property/accessor decorator_, makes a property observable on every instance and enforces that values are not undefined or null when read (throws TypeError) and that only instances of Object can be set, which are sealed automatically [requires ES5+ target] [decorator] */
function observable_seal(target, key, descriptor) {
    return observable(target, key, descriptor, function (value) {
        if (value === undefined || value === null)
            throw new TypeError(key + " is " + value);
        return value;
    }, function (value) {
        if (!(value instanceof Object))
            throw new Error(key + " must be an Object, not " + value);
        return Object.isSealed(value) ? value : Object.seal(value);
    });
}
exports.observable_seal = observable_seal;
/** _Property/accessor decorator_, makes a property observable on every instance and enforces that values are not undefined or null when read (throws TypeError) and that only instances of Object can be set, which are frozen automatically [requires ES5+ target] [decorator] */
function observable_freeze(target, key, descriptor) {
    return observable(target, key, descriptor, function (value) {
        if (value === undefined || value === null)
            throw new TypeError(key + " is " + value);
        return value;
    }, function (value) {
        if (!(value instanceof Object))
            throw new Error(key + " must be an Object, not " + value);
        return Object.isFrozen(value) ? value : Object.freeze(value);
    });
}
exports.observable_freeze = observable_freeze;
