"use strict";
var Defer_1 = require("./Defer");
var Signal_1 = require("./Signal");
/** Represents a value to be resolved at any time in the future */
var Promise = (function () {
    /** Create a new Promise instance by running the given callback function, which is given callbacks to resolve and/or reject the promise; the promise is also rejected if the callback function throws an exception */
    function Promise(executor) {
        /* @internal */
        this._Resolve = Signal_1.defineSignal();
        /* @internal */
        this._Reject = Signal_1.defineSignal();
        if (typeof executor === "function")
            this._resolveWith(executor);
    }
    /** Delay the execution of a callback but return a promise for its result */
    Promise.delay = function (f, ms, args) {
        var result = new Promise();
        // set a timer and run all deferred functions right away (e.g. .then(...))
        setTimeout(function () {
            result._resolveWith(function (r) { return r(f.apply(undefined, args)); });
            Defer_1.yieldAll();
        }, ms);
        return result;
    };
    /** Return a promise that will be resolved after a delay */
    Promise.sleep = function (ms, value) {
        return Promise.delay(function () { return value; }, ms);
    };
    /** Defer the execution of a callback but return a promise for its result */
    Promise.defer = function (f, args) {
        var result = new Promise();
        Defer_1.defer(function () { result._resolveWith(function (r) { return r(f.apply(undefined, args)); }); });
        return result;
    };
    /** Executes given callback, providing it with a Node.js-style (err, result) => {...} handler that immediately resolves or rejects the resulting promise when called; (use as e.g. `Promise.fn(f => fs.readFile('/etc/passwd', f)).then(...))` */
    Promise.fn = function (callback) {
        return new Promise(function (resolve, reject) {
            callback(function (err, result) {
                if (err)
                    reject(err);
                else
                    resolve(result);
            });
        });
    };
    /** Return a resolved promise */
    Promise.resolve = function (value) {
        return new Promise()._resolve(value);
    };
    /** Return a rejected promise */
    Promise.reject = function (error) {
        return new Promise()._reject(error);
    };
    /** Return a promise that is fulfilled when all given promises are fulfilled and is immediately rejected when one of the promises is rejected */
    Promise.all = function (promises) {
        var result = new Promise();
        var values = [];
        var left = promises.length;
        // fulfill promise if not rejected before
        function fulfill() {
            !result._isRejected && result._resolve(values);
        }
        // wait for each given promise
        promises.forEach(function (p, i) {
            p.then(function (value) {
                values[i] = value;
                --left || fulfill();
            }, function (error) {
                result._reject(error);
            });
        });
        // fulfill already if no promises given
        !left && fulfill();
        return result;
    };
    /** Return a promise that is resolved or rejected exactly like the first of the given promises that is resolved or rejected */
    Promise.race = function (promises) {
        var result = new Promise();
        promises.forEach(function (p) {
            p.then(function (value) { return result._resolve(value); }, function (error) { return result._reject(error); });
        });
        return result;
    };
    /** Run one of the callbacks as soon as the promise is fulfilled or rejected */
    Promise.prototype.then = function (onFulfilled, onRejected) {
        var _this = this;
        var result = new Promise();
        if (this._isResolved) {
            // if already fulfilled, schedule call to onFulfilled or fulfill now
            if (typeof onFulfilled === "function")
                Defer_1.defer(function () { result._resolveWith(function (r) { return r(onFulfilled(_this._value)); }); });
            else
                result._resolve(this._value);
        }
        else if (this._isRejected) {
            // if already rejected, schedule call to onRejected or reject now
            if (typeof onRejected === "function")
                Defer_1.defer(function () { result._resolveWith(function (r) { return r(onRejected(_this._error)); }); });
            else
                result._reject(this._error);
            // set flag to make sure UnhandledException does not get called async
            this._hndRejection = true;
        }
        else {
            // fulfill result when ready
            if (typeof onFulfilled === "function") {
                this._Resolve._connect(function (v) {
                    return result._resolveWith(function (r) { return r(onFulfilled(v)); });
                });
            }
            else {
                // pass on result if no handler passed in
                this._Resolve._connect(function (v) { return result._resolve(v); });
            }
            // handle rejections
            if (typeof onRejected === "function") {
                this._Reject._connect(function (v) {
                    return result._resolveWith(function (r) { return r(onRejected(v)); });
                });
            }
            else {
                // pass on rejections if no handler passed in
                this._Reject._connect(function (v) { return result._reject(v); });
            }
        }
        // always return a promise with the result of onFulfilled OR onRejected
        return result;
    };
    /** Alternative method to catch rejections */
    Promise.prototype.catch = function (onRejected) {
        return this.then(undefined, onRejected);
    };
    /** @internal Return current status: null if unfulfilled, object with value property (if resolved), and/or error property (if rejected) */
    Promise.prototype._getStatus = function () {
        if (this._isResolved)
            return { value: this._value };
        if (this._isRejected)
            return { rejected: true, error: this._error };
        return null;
    };
    /** @internal Resolve the promise with a value, or (future) result of a promise */
    Promise.prototype._resolve = function (value, resolving) {
        var _this = this;
        if (!this._isResolved && !this._isRejected &&
            (!this._resolving || resolving === this._resolving)) {
            if (value === this)
                throw new TypeError("Recursive promise found");
            try {
                var thenFn = value && value.then;
                if (typeof thenFn === "function" &&
                    ((typeof value === "object") || (typeof value === "function"))) {
                    // wait for promise to be resolved and recurse
                    resolving = this._resolving = {};
                    thenFn.call(value, function (v) { _this._resolve(v, resolving); }, function (e) { _this._reject(e, resolving); });
                }
                else {
                    // set status and schedule handler chain (if any)
                    this._isResolved = true;
                    this._resolving = null;
                    this._value = value;
                    this._Resolve.emit(value);
                    // remove signal references to help garbage collection
                    delete this._Resolve;
                    delete this._Reject;
                }
            }
            catch (e) {
                // oops, caught an exception, now reject this promise
                this._reject(e, resolving);
            }
        }
        return this;
    };
    /** @internal Run a function that may resolve or reject the promise */
    Promise.prototype._resolveWith = function (executor) {
        var _this = this;
        try {
            // call resolver function with callback for resolving this promise
            executor(function (v) { _this._resolve(v); }, function (e) { _this._reject(e); });
        }
        catch (e) {
            // oops, caught an exception, now reject this promise
            this._reject(e);
        }
        return this;
    };
    /** @internal Reject the promise because an error occurred */
    Promise.prototype._reject = function (error, resolving) {
        var _this = this;
        // set status and schedule handler chain (if any)
        if (!this._isResolved && !this._isRejected &&
            (!this._resolving || resolving === this._resolving)) {
            this._isRejected = true;
            this._error = error;
            if (this._Reject.isConnected()) {
                // emit signal which is handled by onRejected handler(s)
                this._Reject.emit(error);
                // remove signal references to help garbage collection
                delete this._Resolve;
                delete this._Reject;
            }
            else {
                // emit signal async if still not handled
                // (wait for chained calls on already-rejected promise)
                Defer_1.defer(function () {
                    _this._hndRejection ||
                        Signal_1.UnhandledException.emit(error);
                });
            }
        }
        return this;
    };
    return Promise;
}());
exports.Promise = Promise;
/** Return a promise that will be resolved after a delay */
function sleep(ms, value) {
    return Promise.delay(function () { return value; }, ms);
}
exports.sleep = sleep;
