import { IPromise_Thenable } from "./Promise";
import { ObservableObject } from "./ObservableObject";
import { ObservableArray } from "./ObservableArray";
/** Represents an observable value (with the value itself in `.value`) */
export declare class ObservableValue<T> {
    /** Encapsulate given value as an ObservableValue */
    static fromValue<T>(value: T): ObservableValue<T>;
    /** Encapsulate given promised value as an ObservableValue; the observed value will start out as `undefined` but then changes to the `Promise` result when resolved; or if promise was rejected, hangs on to the error and throws it when retrieving the observable value */
    static fromPromise<T>(valuePromise: IPromise_Thenable<T>): ObservableValue<T>;
    /** Create a new (single) observed value using optional getter and setter functions; note that getters *should* be pure functions without side effects, creating or setting other observables from the getter will result in an error; getter is not called immediately but only when being subscribed to or when obtaining the value itself (from `.value`) */
    constructor(getter?: () => T, setter?: (value: T) => void);
    /** Set a getter function for the observable value, which should return a current value, or another ObservableValue instance, or set `.value` directly; note that getters *should* be pure functions without side effects, creating or setting other observables from the getter will result in an error; the getter is only invoked (asynchronously) if a value had been set previously and needs to be updated; reading `.value` from the getter results in the value previously set, no recursion occurs; returns this */
    getter(f: () => T): this;
    /** Set a setter function for the observable value; setting `.value` directly from the setter results in changing the underlying observable value (which can also be read by the getter, or the setter itself), no recursion occurs; returns this */
    setter(f: (value: T) => void): this;
    /** Observable value, (re-) evaluated only if necessary; when set to an `ObservableValue` instance, this property returns that instance's value, until this property is set to another value (except if the `ObservableValue` has a setter, which is called first); plain Array values are turned into `ObservableArray` instances, and plain Object instances into `ObservableObject` instances --- unless `.shallow` is set to true */
    value: T | undefined;
    /** Returns .value (observable if used within an observable getter) */
    valueOf(): T | undefined;
    /** Returns .value as a string (observable if used within an observable getter) */
    toString(): string;
    /** Returns last value set, does not re-evaluate and/or add dependency */
    getLastValue(): T | undefined;
    /** Transform this observable value using given function, into a new ObservableValue instance; note that the transformation function is not necessarily invoked after each value change, if the new observable is not subscribed to and/or multiple changes occur (asynchronously) before the transformation is evaluated; observables used by the given function itself are not automatically subscribed to (wrap given function in observe(...) to observe dependencies as well) */
    map<U>(callback: (value: T) => U): ObservableValue<U>;
    /** Start listening for changes to this observable value and all of its dependencies asynchronously; returns this */
    subscribe(): this;
    /** Start listening for changes to this observable value and all of its dependencies asynchronously, and schedule given function for every new value (including the current value); returns this */
    subscribe(callback: (value: T) => any): this;
    /** Start listening for changes to this observable value and all of its dependencies asynchronously, and return a promise for the next (different) value of the observable; stops subscribing immediately after the promise is resolved */
    next(): PromiseLike<T>;
    /** True if this observable value is currently subscribed to, either directly or from dependent observables */
    readonly subscribed: boolean;
    /** True if this observable value is writable (not only a getter defined) */
    readonly writable: boolean;
    /** Set to true to stop this observablue value instance from converting arrays and objects to observables, and reading values from observable value instances */
    shallow?: boolean;
    /** Clear value and remove getter/setter functions, unsubscribe from dependencies */
    clear(): void;
    /** Manually invoke the getter function synchronously, if any; can be used only if not currently evaluating another observable value (otherwise, wrap in a call to `unobserved`); calling this method is normally not necessary, and should only be used if external factors outside of observables change in such a way that the result of the getter function changes; if so, then subsequent retrieval of `.value` results in the new value, and dependent observable values are scheduled to update automatically (asynchronously) */
    update(): void;
}
/** Create a (single) observable that holds the return value of the given function, re-evaluated (when subscribed to) whenever one of the observable values used in the getter function change, *or* (also when not subscribed to) when `ObservableValue#value` is read after one of the observables used in the getter function (may have) changed its value; note that getters *should* be pure functions without side effects, creating or setting other observables from the getter will result in an error */
export declare function observe<T>(f: () => T): ObservableValue<T>;
/** Encapsulate given promise as an observable value that is set when the promise resolves; if the promise is already resolved, the observable value is set to the promise's value immediately; if or when the promise is rejected, the error is stored and will be thrown when trying to obtain the observable value */
export declare function observe<T>(promise: IPromise_Thenable<T>): ObservableValue<T>;
/** Returns a new ObservableArray with elements copied from given array; (same as `ObservableArray.fromArray`; does not observe values of a given `ObservableArray`, see `observeArray` instead) */
export declare function observe<T>(obj: Array<T>): ObservableArray<T>;
/** Returns a new ObservableObject with properties copied from given object; (same as new ObservableObject(...) but with a strongly typed return value) */
export declare function observe<T extends {}>(obj: T): T & ObservableObject;
/** Invoke given function without recording dependencies on currently evaluating observable values; returns the function's return value */
export declare function unobserved<T>(f: ((...args: any[]) => T), ...args: any[]): T;
