import { ObservableValue } from "./Observable";
/** Encapsulates `Array` with observable properties; the result works exactly like a regular array, but setting elements outside the bounds of the array (>= length) does NOT work: length must be set first */
export declare class ObservableArray<T> {
    /** Create an `ObservableArray` out of a regular Array */
    static fromArray<T>(array: Array<T>): ObservableArray<T>;
    /** Create an `ObservableArray` that takes array elements from the array in given `ObservableValue`, or the value itself as a single element if it is not an array, or an empty array if the value is null or undefined; changes in array elements and/or observable value are reflected asynchronously */
    static fromObservableValue<T>(observableValue: ObservableValue<T[] | T>): ObservableArray<T>;
    /** Create a new empty `ObservableArray` instance */
    constructor();
    /** Create a read-only `ObservableArray` with each value of the original array mapped to the result of the given getter function; observable values used in the map function are not observed (like `ObservableValue#map`); the getter is called for combinations of value and index (i.e. deleting a value in the middle of the source array will trigger changes for all elements after it; if the index is not important then use `.mapAsyncValues` instead); the resulting array length changes along with the original array length */
    mapAsync<U>(callback: (value: T, index: number, array: T[]) => (ObservableValue<U> | U), thisArg?: any): ObservableArray<U>;
    /** Create a read-only ObservableArray with each value of the original array mapped to the result of the given function; observable values used in the map function are not observed (like ObservableValue map method); the resulting array length changes along with the original array length; this method is slightly more expensive than mapAsync for larger arrays, especially in non-ES6 environments, but is overall more efficient because it avoids unnecessary callbacks when subscribed to */
    mapAsyncValues<U>(callbackfn: (value: T) => (ObservableValue<U> | U)): ObservableArray<U>;
    /** Represent observable arrays as regular JSON arrays */
    toJSON(): T[];
}
/** Type definition to declare Array methods mixin */
export interface ObservableArray<T> extends Array<T> {
}
/** Returns an ObservableArray instance based on the result of given function: if the returned value is an array, then all elements are copied to the result; if the returned value is an observable array, all elements are proxied; if the returned value is not an array, the result contains this value as a single element; if the returned value is null or undefined, the result will be an empty array; changes in values/elements are reflected asynchronously */
export declare function observeArray<T>(f: () => (T[] | T)): ObservableArray<T>;
