import { Signal, CustomSignalClass } from "./Signal";
import { ObservableValue } from "./Observable";
/** Represents an object with (some) observable members; may be extended into a derived class, or use `makeObjectObservable` function to mix into any object [requires ES5+ target] */
export declare class ObservableObject {
    /** Create an object with (some) observable members */
    constructor();
    /** Returns true if property with given name is observable */
    hasObservableProperty(name: string | number): boolean;
    /** Signal that is emitted when any property changes (but does not subscribe to any properties, i.e. only changes to properties already subscribed to, and properties with plain values [not getters] will trigger this signal) */
    PropertyChange: CustomSignalClass<string, typeof ObservableObject.PropertyChangeSignal>;
}
export declare namespace ObservableObject {
    /** Base class for signals that are emitted when any observed property is modified on an ObservableObject instance, with the name of the property as a parameter */
    class PropertyChangeSignal extends Signal<string> {
        /** The ObservableObject instance that this signal is used on */
        static target: ObservableObject;
    }
}
/** Returns a new `ObservableObject` with properties copied from given object; `ObservableValue` instances are referenced directly (including getter and setter), Array properties are turned into `ObservableArray` instances, and plain Object instances into `ObservableObject` instances; also seals the new instance if the object parameter was sealed, and freezes it if the object parameter was frozen*/
export declare function makeObjectObservable<T extends {}>(obj: T): T & ObservableObject;
/** Make the property with given name observable; use given observable value instance, if any, otherwise use a new observable value instance that holds the current property value and/or uses the current property getter and setter [requires ES5+ target] */
export declare function makePropertyObservable(obj: {}, member: string | number, instance?: ObservableValue<any>): void;
/** Delete the property with given name and clear its observable value */
export declare function deleteObservableProperty(obj: {}, member: string | number): void;
/** Returns true if property with given name is defined and is observable */
export declare function isObservableProperty(obj: {}, member: string | number): boolean;
/** _Property/accessor decorator_, makes a property observable on every instance [requires ES5+ target] [decorator] */
export declare function observable(target: Object, key: string, descriptor?: PropertyDescriptor): any;
/** _Property/accessor decorator_, makes a property observable on every instance and converts values to strings when read (empty string for null/undefined/NaN) [requires ES5+ target] [decorator] */
export declare function observable_string(target: Object, key: string, descriptor?: PropertyDescriptor): any;
/** _Property/accessor decorator_, makes a property observable on every instance and converts values to numbers using Number(...) when read [requires ES5+ target] [decorator] */
export declare function observable_number(target: Object, key: string, descriptor?: PropertyDescriptor): any;
/** _Property/accessor decorator_, makes a property observable on every instance, as a shallow observable value (i.e. does not read values from observable values assigned to this property; and does not convert arrays to observable arrays nor objects to observable objects) [requires ES5+ target] [decorator] */
export declare function observable_shallow(target: Object, key: string, descriptor?: PropertyDescriptor): any;
/** _Property/accessor decorator_, makes a property observable on every instance and enforces that values are not undefined or null when read (throws TypeError) and that undefined/null values cannot be set (also throws TypeError) [requires ES5+ target] [decorator] */
export declare function observable_not_null(target: Object, key: string, descriptor?: PropertyDescriptor): any;
/** _Property/accessor decorator_, makes a property observable on every instance and enforces that values are not undefined or null when read (throws TypeError) and that only instances of Object can be set, which are sealed automatically [requires ES5+ target] [decorator] */
export declare function observable_seal(target: Object, key: string, descriptor?: PropertyDescriptor): any;
/** _Property/accessor decorator_, makes a property observable on every instance and enforces that values are not undefined or null when read (throws TypeError) and that only instances of Object can be set, which are frozen automatically [requires ES5+ target] [decorator] */
export declare function observable_freeze(target: Object, key: string, descriptor?: PropertyDescriptor): any;
