/** Basic interface of Promise type, compatible with `PromiseLike` (ES6 targeted TypeScript) but also including .catch(...) method */
export interface IPromise<T> {
    /** Run one of the callbacks as soon as the promise is fulfilled or rejected */
    then<O>(onFulfilled?: (value: T) => (IPromise<O> | O), onRejected?: (error: Error) => (IPromise<O> | O)): IPromise<O>;
    /** Run one of the callbacks as soon as the promise is fulfilled or rejected */
    then<O>(onFulfilled?: (value: T) => (IPromise<O> | O), onRejected?: (error: Error) => void): IPromise<O>;
    /** Alternative method to catch rejections */
    catch<O>(onRejected?: (error: Error) => (IPromise<O> | O)): IPromise<O>;
    /** Alternative method to catch rejections */
    catch<O>(onRejected?: (error: Error) => void): IPromise<O>;
}
/** Basic interface of Promise type, compatible with `PromiseLike` (ES6 targeted TypeScript) without .catch(...) method */
export interface IPromise_Thenable<T> {
    /** Run one of the callbacks as soon as the promise is fulfilled or rejected */
    then<O>(onFulfilled?: (value: T) => (IPromise_Thenable<O> | O), onRejected?: (error: any) => (IPromise_Thenable<O> | O)): IPromise_Thenable<O>;
    /** Run one of the callbacks as soon as the promise is fulfilled or rejected */
    then<O>(onFulfilled?: (value: T) => (IPromise_Thenable<O> | O), onRejected?: (error: any) => void): IPromise_Thenable<O>;
}
/** Represents a value to be resolved at any time in the future */
export declare class Promise<T> implements IPromise<T> {
    /** Delay the execution of a callback but return a promise for its result */
    static delay<T>(f: (...args: any[]) => T, ms: number, args?: any[]): IPromise<T>;
    /** Return a promise that will be resolved after a delay */
    static sleep<PromiseT>(ms: number, value?: PromiseT): IPromise<PromiseT | undefined>;
    /** Defer the execution of a callback but return a promise for its result */
    static defer<T>(f: (...args: any[]) => T, args?: any[]): IPromise<T>;
    /** Executes given callback, providing it with a Node.js-style (err, result) => {...} handler that immediately resolves or rejects the resulting promise when called; (use as e.g. `Promise.fn(f => fs.readFile('/etc/passwd', f)).then(...))` */
    static fn<T>(callback: (f: (err: any, result: T) => void) => void): Promise<T>;
    /** Return a resolved promise */
    static resolve<T>(value: T): Promise<T>;
    /** Return a rejected promise */
    static reject(error: Error): Promise<{}>;
    /** Return a promise that is fulfilled when all given promises are fulfilled and is immediately rejected when one of the promises is rejected */
    static all<ValueT>(promises: IPromise_Thenable<ValueT>[]): Promise<ValueT[]>;
    /** Return a promise that is resolved or rejected exactly like the first of the given promises that is resolved or rejected */
    static race<T>(promises: IPromise_Thenable<T>[]): Promise<T>;
    /** Create a new Promise instance by running the given callback function, which is given callbacks to resolve and/or reject the promise; the promise is also rejected if the callback function throws an exception */
    constructor(executor?: (resolve: (value: T | IPromise<T>) => void, reject: (error: Error) => void) => void);
    /** Run one of the callbacks as soon as the promise is fulfilled or rejected */
    then<O>(onFulfilled?: (value: T) => (IPromise<O> | O), onRejected?: (error: Error) => (IPromise<O> | O)): Promise<O>;
    /** Alternative method to catch rejections */
    catch<O>(onRejected?: (error: Error) => (IPromise<O> | O)): Promise<O>;
}
/** Return a promise that will be resolved after a delay */
export declare function sleep<PromiseT>(ms: number, value?: PromiseT): IPromise<PromiseT | undefined>;
