import { IPromise } from "./Promise";
/** A static class for a signal that can be used to trigger one or more handlers asynchronously, with a base class other than Signal itself (passed to connected handlers); use `defineSignal` to define classes to match this type */
export interface CustomSignalClass<DataT, BaseClassT extends typeof Signal> {
    /** Instantiate a signal with given value, ready to be emitted */
    new (data: DataT): Signal<DataT>;
    /** Emit a signal with given value */
    (data?: DataT): void;
    /** Add a handler to be invoked when this signal is emitted; returns an encapsulation of the connection with a disconnect method */
    connect(callback: (data: DataT, source: BaseClassT) => any): SignalConnection;
    /** Add a handler to be invoked when this signal is emitted: a method with given name on given target object (evaluated only when needed); returns an encapsulation of the connection with a disconnect method */
    connect(method: string, target: {}): SignalConnection;
    /** Add a handler to be invoked the next time this signal is emitted */
    connectOnce(callback: (data: DataT, source: BaseClassT) => any): void;
    /** Add a handler to be invoked the next time this signal is emitted: a method with given name on given target object (evaluated only when needed) */
    connectOnce(method: string, target: {}): void;
    /** Remove all handlers */
    disconnectAll(): void;
    /** Returns true if this signal has any handlers connected to it */
    isConnected(): boolean;
}
/** A static class for a signal that can be used to trigger one or more handlers asynchronously; use `defineSignal` to define classes to match this type */
export interface SignalClass<DataT> extends CustomSignalClass<DataT, typeof Signal> {
    /** Add a handler to be invoked when this signal is emitted; returns an encapsulation of the connection with a disconnect method */
    connect(callback: (data: DataT) => any): SignalConnection;
    /** Add a handler to be invoked when this signal is emitted: a method with given name on given target object (evaluated only when needed); returns an encapsulation of the connection with a disconnect method */
    connect(method: string, target: {}): SignalConnection;
    /** Add a handler to be invoked the next time this signal is emitted */
    connectOnce(callback: (data: DataT) => any): void;
    /** Add a handler to be invoked the next time this signal is emitted: a method with given name on given target object (evaluated only when needed) */
    connectOnce(method: string, target: {}): void;
}
/** Encapsulates a handler connected to a signal */
export interface SignalConnection {
    /** Disconnect the connected handler from this signal */
    disconnect(): void;
    /** True if connection has been disconnected */
    disconnected: boolean;
}
/** A static class for a signal that can be used to trigger one or more handlers asynchronously (use `defineSignal` to create classes) */
export declare abstract class Signal<T> {
    /** Add a handler to be invoked when this signal is emitted; returns an encapsulation of the connection with a disconnect method */
    static connect(callback: (data: any, source: typeof Signal) => any): SignalConnection;
    /** Add a handler to be invoked when this signal is emitted: a method with given name on given target object (evaluated only when needed); returns an encapsulation of the connection with a disconnect method */
    static connect(method: string, target: {}): SignalConnection;
    /** Add a handler to be invoked the next time this signal is emitted */
    static connectOnce(callback: (data: any, source: typeof Signal) => any): any;
    /** Add a handler to be invoked the next time this signal is emitted: a method with given name on given target object (evaluated only when needed) */
    static connectOnce(method: string, target: {}): any;
    /** Remove all handlers; returns this */
    static disconnectAll(): typeof Signal;
    /** Returns true if this signal has any handlers */
    static isConnected(): boolean;
    /** Static method that is called synchronously when a first handler is connected (and when a new handler is connected after all had been disconnected); override this in a signal base class, e.g. to add a delayed initialization method */
    protected static onHandlerConnected?: () => void;
    /** Static method that is called synchronously when no more handlers are connected; override this in a signal base class, e.g. to add a delayed deallocation method */
    protected static onHandlersDisconnected?: () => void;
    /** Create a signal with given data, ready to be emitted */
    constructor(data: T);
    /** Invoke all handlers and capture promises of their return values in .results (unless argument is true); works only once; returns this */
    emit(noResults?: boolean): this;
    /** Invoke given callback(s) with return values of all handlers, or any exception that occurred during execution of all handlers; returns a Promise that resolves to the return value of the callback(s) itself */
    then<O>(onFulfilled?: (results: any[]) => O, onRejected?: (error: Error) => O): IPromise<O>;
    /** Catch errors that occur during execution of all handlers; returns a Promise that resolves to the return value of the callback itself */
    catch<O>(onRejected?: (error: Error) => O): IPromise<O>;
    /** Array of Promises that resolve to return values of all handlers */
    readonly results: IPromise<any>[];
}
/** Create a new `Signal` class with given data type */
export declare function defineSignal<DataT>(): SignalClass<DataT>;
/** Create a new Signal class with given base class, and set given (static) properties on the created class; the created signal class derives from the base class */
export declare function defineSignal<BaseClassT extends typeof Signal, DataT>(Base: BaseClassT & {
    new (data: DataT): Signal<DataT>;
}, properties?: {}): CustomSignalClass<DataT, BaseClassT>;
/** Signal that is triggered with exceptions that were unhandled during async execution; a custom handler may be added here, the default handler just writes a warning message to the console; to disable this behavior for specific errors, set error.message to a blank string */
export declare const UnhandledException: SignalClass<Error>;
